# pylint: disable=wildcard-import
"""
SMOT: Single-Shot Multi Object Tracking
https://arxiv.org/abs/2010.16031
"""
from __future__ import absolute_import
from .smot_tracker import *
from .tracktors import *
